/* win32zip.c
 Copyright (C) 1990-1996 Mark Adler, Richard B. Wales, Jean-loup Gailly,
 Kai Uwe Rommel, Onno van der Linden and Igor Mandrichenko.
 This version modified by Eric W. Engler for use with Delphi ZIP.
*/

#ifndef UTIL    /* this file contains nothing used by UTIL */

#include "zip.h"
#include "win32zip.h"

#include <direct.h>     /* for rmdir() */
#include <time.h>
#include <sys/utime.h>
#include <windows.h>    /* for findfirst/findnext stuff */
#include <io.h>

char *GetLongPathEA OF((char *name));

#define MATCH         dosmatch   /* EWE CHG: shmatch */

#define PAD           0
#define PATH_END      '/'
#define HIDD_SYS_BITS (FILE_ATTRIBUTE_HIDDEN | FILE_ATTRIBUTE_SYSTEM)

typedef struct zdirent {
  ush    d_date, d_time;
  ulg    d_size;
  char   d_attr;
  char   d_name[MAX_PATH];
  int    d_first;
  HANDLE d_hFindFile;
} zDIR;

/* Local functions */
local zDIR           *Opendir  OF((const char *));
local struct zdirent *Readdir  OF((zDIR *));
local void            Closedir OF((zDIR *));
local char           *readd    OF((zDIR *));

/* Module level variables */
extern char *label /* = NULL */ ;       /* defined in fileio.c */
local ulg label_time = 0;
local ulg label_mode = 0;
local time_t label_utim = 0;

/* Module level constants */
local const char wild_match_all[] = "*.*";

/* Start searching for files in the MSDOS directory "n" */
/* Return a pointer to a zDIR struct. */
local zDIR *Opendir(n)
const char *n;          /* directory to open */
{
  zDIR *d;              /* malloc'd return value */
  char *p;              /* malloc'd temporary string */
  char *q;
  WIN32_FIND_DATA fd;

  sprintf(ewemsg, "Opendir win32zip.c, path passed in =%s", n);
  diag(ewemsg);

  if ((d = (zDIR *)malloc(sizeof(zDIR))) == NULL ||
    (p = malloc(strlen(n) + 5)) == NULL) {
       if (d != NULL) 
          free((zvoid *)d);
       return NULL;
  }
  /* d points to our return zDIR struct,
     p points to a temp buf where we'll play with pathname */
  strcpy(p, n);

  q = p + strlen(p);
  /* q points to null byte */

  /*    q - p is length of pathname */
  /*  *(q - 1) is last byte of pathname */

  if ((q - p) > 0 && *(q - 1) != '/')
    *q++ = '/';  // add a trailing / if not already there */

  // the following is effectively a strcat 
  strcpy(q, wild_match_all); /*  "*.*"  MS-DOS match-all spec */

  //sprintf(ewemsg, "doing findfirst with spec %s", p);
  //diag(ewemsg);

  d->d_hFindFile = FindFirstFile(p, &fd);
  free((zvoid *)p);

  if (d->d_hFindFile == INVALID_HANDLE_VALUE)
  {
    free((zvoid *)d);
    return NULL;
  }

  strcpy(d->d_name, fd.cFileName);
  //sprintf(ewemsg, "first entry = %s",d->d_name);
  //diag(ewemsg);

  d->d_attr = (unsigned char)fd.dwFileAttributes;
  d->d_first = 1;
  return d;
}

/* Read the next directory entry. Pass in the zDIR ptr obtained from OpenDir.
   Return a pointer to it (zdirent struct *),
   or NULL if end of directory. */
local struct zdirent *Readdir(d)
zDIR *d;    /* directory stream to read from */
{
  if (d->d_first)    // This tells us if we just did opendir, and still
     d->d_first = 0; // have it's spec ready to return to caller.
  else
  {
    WIN32_FIND_DATA fd;

    if (!FindNextFile(d->d_hFindFile, &fd))
        return NULL;
    strcpy(d->d_name, fd.cFileName);
    //sprintf(ewemsg, "NEXT ENTRY = %s",d->d_name);
    //diag(ewemsg);
    d->d_attr = (unsigned char)fd.dwFileAttributes;
  }
  return (struct zdirent *)d;
}

/* Close a directory.  Pass in the zDIR ptr received from OpenDir */
local void Closedir(d)
zDIR *d;              
{
  FindClose(d->d_hFindFile);
  free((zvoid *)d);
}

/* Read next "normal" directory entry.  Hidden and system files are
   skipped.  Pass in the zDIR ptr received from OpenDir.  Return a 
   pointer to the zdirent struct for the next name in the directory, 
   or NULL if no more entries or an error occurs. */
local char *readd(d)
zDIR *d;
{
  struct zdirent *e;

  //diag("readd win32zip.c");

  /* do/while loop to read all dir entries until next normal entry is found */
  do
    e = Readdir(d);
  while (!hidden_files && e && (e->d_attr & HIDD_SYS_BITS) && (!global_abort_sw));

  if (global_abort_sw) 
     return NULL;
  
  if (e == NULL)
     return (char *)NULL;
  else
     return e->d_name;
}

/* If not in exclude mode, expand the pattern based on the contents
   of the file system.  Return an error code in the ZE_ class. 
   This function is used while gathering filenames to be added or updated */
int wild(w)
char *w;                /* path/pattern to match */
{
  zDIR *d;              /* stream for reading directory */
  char *e;              /* name found in directory */
  int r;                /* temporary variable */
  char *n;              /* constructed name from directory */
  int f;                /* true if there was a match */
  char *a;              /* alloc'ed space for name */
  char *p;              /* path */
  char *q;              /* name */
  char v[5];            /* space for device current directory */

  sprintf(ewemsg,"in wild of win32zip.c, recurse=%d", recurse);
  diag(ewemsg);

  /* Allocate and copy pattern */
  if ((p = a = malloc(strlen(w) + 1)) == NULL)
     return ZE_MEM;
  strcpy(p, w);

  /* change all \ to /  */
  for (q = p; *q; q++)
     if (*q == '\\')
        *q = '/';

  /* Only fname (not path) can have special matching characters */
  if ((q = isshexp(p)) != NULL &&
    (strrchr(q, '/') != NULL || strrchr(q, ':') != NULL))
  {
    diag("path had illegal chars");
    free((zvoid *)a);
    return ZE_PARMS;
  }

  /* Separate path and name into p and q */
  if ((q = strrchr(p, '/')) != NULL && (q == p || q[-1] != ':'))
  {
    *q++ = '\0';             
    if (*p == '\0')          /* if path is just / */
      p = strcpy(v, "/.");
  }
  else if ((q = strrchr(p, ':')) != NULL)
  {                                 /* has device and no or root path */
    *q++ = '\0';
    p = strcat(strcpy(v, p), ":");  /* copy device as path */
    if (*q == '/')                  /* -> device:/., name */
    {
      strcat(p, "/");
      q++;
    }
    strcat(p, ".");
  }
  else if (recurse && ((strcmp(p, ".") == 0) || (strcmp(p, "..") == 0)))
  { /* current or parent directory */
    /* I can't understand Mark's code so I am adding a hack here to get
     * "zip -r foo ." to work. Allow the dubious "zip -r foo .." but
     * reject "zip -rm foo ..".  "dispose" means wipe out source path.
     */
    if (dispose && strcmp(p, "..") == 0)
       ziperr(ZE_PARMS);
    q = (char *)wild_match_all;
  }
  else   /* no path or device */
  {
    q = p;
    p = strcpy(v, ".");
  }
  if (recurse && *q == '\0') 
     q = (char *)wild_match_all;
  /* take out a possibly redundant dir name of "." */
  if ((r = strlen(p)) > 1 &&
   (strcmp(p + r - 2, ":.") == 0 || strcmp(p + r - 2, "/.") == 0))
      *(p + r - 1) = '\0';


  //sprintf(ewemsg, "at break up place in wild: p=%s  q=%s", p, q);
  //diag(ewemsg);
  // EWE qqq good place to break up this function 
  // except for this near end:  free((zvoid *)a);


  /* Now that we have a dir spec, along with an fspec, we'll look
     thru the dir specified to see if there's any matches against
     the fspec */
  if ((d = Opendir(p)) == NULL)
  {
    free((zvoid *)a);
    diag("can't open dir");
    return ZE_MISS;
  }

  f = 0;
  /* In the while loop, f is "sticky".  It will stay as 1 if there
     is at least 1 good filename that matches the spec */
  while ((e = readd(d)) != NULL) {
    if (global_abort_sw) 
       return ZE_MISS;
    /* if e is NOT '.' or '..', and it does match fspec...*/
    if (strcmp(e, ".") && strcmp(e, "..") && MATCH(q, e))
    {
      /* we matched fspec, and entry is not '.' or '..' */
      f = 1;
      if (strcmp(p, ".") == 0) {  
         /* path is just '.', so don't even feed it to procname */
         /* "procname" Process: 
            if arg is dir then CD to it, else add file to list */
         r = procname(e);         
         if (r != ZE_OK) {
            f = 0;  /* error - set normally sticky f back to 0 */
            break;
         }
      }
      else {
        /* get a new tmp buffer to hold the path */
        if ((n = malloc(strlen(p) + strlen(e) + 2)) == NULL)
        {
          free((zvoid *)a);
          Closedir(d);
          return ZE_MEM;
        }
        /* copy dir path to new buffer */
        strcpy(n, p);
        /* ensure that path ends in '/' or ':' */
        r = strlen(n) - 1;
        if (n[r] != '/' && n[r] != ':')
           strcat(n, "/");
        /* Put name of current file from dir onto the path, then
           process it.  
           "procname" Process: 
           if arg is dir then CD to it, else add file to list */
        r = procname(strcat(n, e));
        free((zvoid *)n);
        if (r != ZE_OK) {
           f = 0;  /* error - set normally sticky f back to 0 */
           break;
        }
      } /* end else */
    } /* end "if (strcmp..." */
  } /* end while */

  Closedir(d);
  //sprintf(ewemsg,"before exiting wild, f=%d (1 is good)", f);
  //diag(ewemsg);
  free((zvoid *)a);
  return f ? ZE_OK : ZE_MISS;
}

/* Process a name or wildcard expression to operate on (or exclude). 
   Return an error code in the ZE_ class. */
int procname(n)
char *n;                /* name to process */
{
  char *a;              /* path and name for recursion */
  zDIR *d;              /* directory stream from opendir() */
  char *e;              /* pointer to name from readd() */
  int m;                /* matched flag */
  char *p;              /* path for recursion */
  struct stat s;        /* result of stat() */
  struct zlist far *z;  /* steps through zfiles list */

 // sprintf(ewemsg,"in procname(win32zip.c)  name=>%s<=  recurse=%d", n, recurse);
 // diag(ewemsg);

  m = 1;  /* set dflt for success indicator (0=success) */

  if (*n == '\0')
     return ZE_MISS;
  if (global_abort_sw) 
     return ZE_MISS;

  /* LSSTAT returns 0 if it's arg is any kind of file (even a dir). */
  /* isshexp returns true if a wildcard symbol is found, or
     NULL if none were found -- isshexp is in util.c */
  //sprintf(ewemsg,"LSSTAT of %s returns %d", n, LSSTAT(n, &s));
  //diag(ewemsg);

  if ( LSSTAT(n, &s) || (isshexp(n) != NULL))
  {
    //diag("not a file or dir - 'n' must be a wildcard fspec");

    // Upon finding a wildcard fspec, we first need to mark entries in 
    // the "zfiles" list that are included by the wildcard.  Later, 
    // we'll collect new files to add that aren't already in the archive.

    /* convert the "external" (native) filename to an internal
       ZIP-archive-compatible filename. */
    p = ex2in(n, 0, (int *)NULL);   /* shouldn't affect matching chars */

    /* does any file already in the archive match this spec? */
    /* Note that we need the pathname and filename together for this */
    for (z = zfiles; z != NULL; z = z->nxt) {
       if (MATCH(p, z->zname))
       {
          /* name is in archive - mark it for updating */
          z->mark = pcount ? filter(z->zname) : 1;
          if ((verbose) && (!z->mark))
             fprintf(mesg, "%scluding %s\n", z->mark ? "in" : "ex", z->name);
          m = 0; /* we had at least one file in the archive that we marked */
       }
    }
    free((zvoid *)p);

    /* returns 1 if no "zfiles" entries were marked,
       0 if at least one entry was marked */
    if (m != 0)
       {
       //diag("returning ZE_MISS from procname");
       return ZE_MISS;
       }
    //diag("returning ZE_OK from procname");
    return ZE_OK;  // ZE_OK = 0
  } /* end of "if (LSSTAT..." */

  /* good name-- add the name if it's a file, recurse if directory */
  //diag("good entry to add, or a dir to go into");

  /* change all \ to / */
  for (p = n; *p; p++)  
     if (*p == '\\')
        *p = '/';
  /* p now points just past string - it isn't useful anymore */

  /* check the status returned by LSSTAT earlier to see if 'n' is a dir */
  if ((s.st_mode & S_IFDIR) == 0) 
  {
    /* it's not a directory - add file to found list */
    diag("spot 1: add file to found list");
    sprintf(ewemsg, "spot 1: put wildcard matching here. n=%s", n);
    diag(ewemsg);
    /* newname (fileio.c) adds to found list. If error m!=0 on return */
    if ((m = newname(n, 0)) != ZE_OK)  
    {
       sprintf(ewemsg,"returning %d from procname after newname call", m);
       diag(ewemsg);
       return m;
    }
  } else {
    /* It is a directory - Add trailing / to the directory name */
    //diag("spot 2");
    if ((p = malloc(strlen(n)+2)) == NULL)
       return ZE_MEM;
    if ((strcmp(n, ".") == 0) || (strcmp(n, "/.") == 0)) {
       *p = '\0';  /* avoid "./" prefix and do not create zip entry */
    } else {
       //diag("spot 3");
       strcpy(p, n);
       a = p + strlen(p);
       if (a[-1] != '/')
          strcpy(a, "/");
       /* newname (fileio.c) adds to found list. If error m!=0 on return */
       if (dirnames && (m = newname(p, 1)) != ZE_OK) {
          free((zvoid *)p);
          sprintf(ewemsg,"returning %d from procname after 2nd newname call", m);
          diag(ewemsg);
          return m;
       }
    }

    if (global_abort_sw) 
       return ZE_MISS;

    /* recurse into directory */
    //diag("spot 4: optional recurse into dir");
    if (recurse && ((d = Opendir(n)) != NULL)) /* Open new dir (like chdir) */
    {
      //diag("good open of dir"); 
      while ((e = readd(d)) != NULL) {
        if (global_abort_sw) {
           ziperr(ZE_ABORT);
           return 0;
        }
        /* e is pointing to the new filename we just read from dir */
        /* ignore dir entries of . and .. */
        if (strcmp(e, ".") && strcmp(e, ".."))
        {
          /* get a new tmp buffer for the path and fname */
          if ((a = malloc(strlen(p) + strlen(e) + 1)) == NULL)
          {
            Closedir(d);
            free((zvoid *)p);
            return ZE_MEM;
          }

          /* form the new dir's pathname followed by the fname just read */
          /* (we need to send in the dir and fname, or it won't be
             detected as a valid file by LSSTAT) */
          strcat(strcpy(a, p), e);

          //diag("DOING RECURSIVE CALL TO PROCNAME FROM PROCNAME");
          if ((m = procname(a)) != ZE_OK)
          {
            /* recursive call failed; return code not ZE_OK */
            if (m == ZE_MISS)
               zipwarn("name not matched: ", a);
            else
               /* unknown error */
               ziperr(m);
          }
          free((zvoid *)a);
        }
      } /* end while */
      Closedir(d);
    } /* end if (spot 4) */

    free((zvoid *)p);
  } /* (s.st_mode & S_IFDIR) == 0) */
  //diag("returning ZE_OK from procname");
  return ZE_OK;
}

/* Convert the external file name to an "internal" file name that
   is valid to store in the ZIP archive, returning the malloc'ed
   string, or NULL if not enough memory. */
char *ex2in(x, isdir, pdosflag)
char *x;                /* external file name */
int isdir;              /* input: x is a directory */
int *pdosflag;          /* output: force MSDOS file attributes? */
{
  char *n;              /* internal file name (malloc'ed) */
  char *t;              /* shortened name */
  int dosflag;

  //sprintf(ewemsg,"in ex2in.  x=%s  isdir=%d", x, isdir);
  //diag(ewemsg);
  dosflag = dosify || IsFileSystemOldFAT(x);
  if (!dosify && use_longname_ea && (t = GetLongPathEA(x)) != NULL)
  {
     sprintf(ewemsg,"replacing short name: %s with long name: %s", x, t);
     diag(ewemsg);
     x = t;
     dosflag = 0;
  }

  /* Find starting point in name before doing malloc */
  t = *x && *(x + 1) == ':' ? x + 2 : x;
  while (*t == '/' || *t == '\\')
     t++;

  /* Make changes, if any, to the copied name (leave original intact) */
  for (n = t; *n; n++)
     if (*n == '\\')
        *n = '/';

  /* This is where the dirname gets stripped if user doesn't want it */
  if (!pathput)
     t = last(t, PATH_END);

  /* Malloc space for internal name and copy it */
  if ((n = malloc(strlen(t) + 1)) == NULL)
     return NULL;
  strcpy(n, t);

  if (dosify)
    msname(n);

  /* set this return flag as needed */
  if (pdosflag)
    *pdosflag = dosflag;

  sprintf(ewemsg,"ex2in returning: %s", n);
  diag(ewemsg);

  /* Returned malloc'ed name */
  return n;
}

/* Convert the "internal" file name (inside the archive) to an 
   external file name, returning the malloc'ed string,
   or NULL if not enough memory. */
char *in2ex(n)
char *n;         /* internal file name */
{
  char *x;       /* external file name */

  //sprintf(ewemsg,"in in2ex. n=%s", n);
  //diag(ewemsg);

  if ((x = malloc(strlen(n) + 1 + PAD)) == NULL)
     return NULL;
  strcpy(x, n);

  if ( !IsFileNameValid(x) )
     ChangeNameForFAT(x);

  sprintf(ewemsg,"in2ex returning: %s", x);
  diag(ewemsg);

  /* return malloc'ed name */
  return x;
}

/* Set last updated and accessed time of file f to the DOS time d. */
/* This is called by dllzip.c to set date/time of the zipfile */
void stamp(f, d)
char *f;                /* name of file to change */
ulg d;                  /* dos-style time to change it to */
{
#if defined(__TURBOC__) && !defined(__BORLANDC__)
  int h;                /* file handle */

  if ((h = open(f, 0)) != -1)
  {
    setftime(h, (struct ftime *)&d);
    close(h);
  }
#else /* !__TURBOC__ */

  struct utimbuf u;     /* argument for utime() */

  /* Convert DOS time to time_t format in u.actime and u.modtime */
  u.actime = u.modtime = dos2unixtime(d);

  /* Set updated and accessed times of f */
  utime(f, &u);
#endif /* ?__TURBOC__ */
}

/* If file *f does not exist, return 0.  Else, return the file's last
   modified date and time as an MSDOS date and time.  The date and
   time is returned in a long with the date most significant to allow
   unsigned integer comparison of absolute times.  Also, if a is not
   a NULL pointer, store the file attributes there, with the high two
   bytes being the Unix attributes, and the low byte being a mapping
   of that to DOS attributes.  If n is not NULL, store the file size
   there.  If t is not NULL, the file's access and modification time
   are stored there as UNIX time_t values.
   If f is "-", use standard input as the file. If f is a device, return
   a file size of -1 */
ulg filetime(f, a, n, t)
char *f;                /* name of file to get info on */
ulg *a;                 /* return value: file attributes */
long *n;                /* return value: file size */
ztimbuf *t;             /* return value: access and modification time */
{
  struct stat s;        /* results of stat() */
  char name[FNMAX];
  int len = strlen(f), isstdin = !strcmp(f, "-");

  if (f == label) {
    if (a != NULL)
      *a = label_mode;
    if (n != NULL)
      *n = -2L; /* convention for a label name */
    if (t != NULL)
      t->actime = t->modtime = label_utim;
    return label_time;
  }
  strcpy(name, f);
  /* not all systems allow stat'ing a file with / appended, so remove it */
  if (name[len - 1] == '/')
     name[len - 1] = '\0';
  /* Accept about any kind of file including directories */
  if (LSSTAT(name, &s) != 0)
     return 0; // error in stat!

  if (a != NULL) {
     *a = ((ulg)s.st_mode << 16) | (isstdin ? 0L : (ulg)GetFileMode(name));
  }
  if (n != NULL)
     *n = (s.st_mode & S_IFMT) == S_IFREG ? s.st_size : -1L;
  if (t != NULL) {
     t->actime = s.st_atime;
     t->modtime = s.st_mtime;
  }
  return unix2dostime((time_t *)&s.st_mtime);
}

/* create extra field and change z->att if desired */
int set_extra_field(z, z_utim)
  struct zlist far *z;
  ztimbuf *z_utim;
{
#ifdef USE_EF_UX_TIME
  if ((z->extra = (char *)malloc(EB_HEADSIZE+EB_UX_MINLEN)) == NULL)
    return ZE_MEM;

  z->extra[0]  = 'U';
  z->extra[1]  = 'X';
  z->extra[2]  = EB_UX_MINLEN;          /* length of data part of e.f. */
  z->extra[3]  = 0;
  z->extra[4]  = (char)(z_utim->actime);
  z->extra[5]  = (char)(z_utim->actime >> 8);
  z->extra[6]  = (char)(z_utim->actime >> 16);
  z->extra[7]  = (char)(z_utim->actime >> 24);
  z->extra[8]  = (char)(z_utim->modtime);
  z->extra[9]  = (char)(z_utim->modtime >> 8);
  z->extra[10] = (char)(z_utim->modtime >> 16);
  z->extra[11] = (char)(z_utim->modtime >> 24);

  z->cext = z->ext = (EB_HEADSIZE+EB_UX_MINLEN);
  z->cextra = z->extra;

  return ZE_OK;
#else /* !USE_EF_UX_TIME */
  return (int)(z-z);
#endif /* ?USE_EF_UX_TIME */
}

/* Delete the directory *d if it is empty, do nothing otherwise.
   Return the result of rmdir(), delete(), or system().
 */
int deletedir(d)
char *d;                /* directory to delete */
{
    return rmdir(d);
}

#endif    /* this file contains nothing used by UTIL */

